
#include <string.h>
#include "pgp.h"
#include "exitpgp.h"
#include "block.h"
#include "randpool.h"

/*************************************************************************/

static char szIDEA[]="IDEA";
static char szTDES[]="3DES";
static char szCAST5[]="CAST5";
static char szBLOWFISH[]="BLOWFISH";
static char szAES128[]="AES128";
static char szAES192[]="AES192";
static char szAES256[]="AES256";
static char szTWOFISH[]="TWOFISH";
static char szOther[12]="";

char * blockCipherName(int Cipher)
{
    switch (Cipher) {
    case IDEA_ALGORITHM_BYTE: return szIDEA;
    case TDES_ALGORITHM_BYTE: return szTDES;
    case CAST5_ALGORITHM_BYTE: return szCAST5;
    case BLOWFISH_ALGORITHM_BYTE: return szBLOWFISH;
    case AES128_ALGORITHM_BYTE: return szAES128;
    case AES192_ALGORITHM_BYTE: return szAES192;
    case AES256_ALGORITHM_BYTE: return szAES256;
    case TWOFISH_ALGORITHM_BYTE: return szTWOFISH;
    default: sprintf(szOther, "%d", Cipher); return szOther;
    }
}

void blockCfbReinit(struct BlockCfbContext *context, byte const *iv)
{
    if (iv)
	    memcpy(context->iv, iv, context->blocksize);
    else
	    fill0(context->iv, context->blocksize);
    context->bufleft = 0;
}

int blockBlockSize(int Cipher)
{
    switch (Cipher) {
    case IDEA_ALGORITHM_BYTE:
    case TDES_ALGORITHM_BYTE:
    case CAST5_ALGORITHM_BYTE:
    case BLOWFISH_ALGORITHM_BYTE:
        return 8;
    case AES128_ALGORITHM_BYTE:
    case AES192_ALGORITHM_BYTE:
    case AES256_ALGORITHM_BYTE:
    case TWOFISH_ALGORITHM_BYTE:
        return 16;
    default:
        return 0;
    }
}

int blockKeySize(int Cipher)
{
    switch (Cipher) {
    case IDEA_ALGORITHM_BYTE:
    case CAST5_ALGORITHM_BYTE:
    case BLOWFISH_ALGORITHM_BYTE:
    case AES128_ALGORITHM_BYTE:
        return 16;
    case TDES_ALGORITHM_BYTE:
    case AES192_ALGORITHM_BYTE:
        return 24;
    case AES256_ALGORITHM_BYTE:
    case TWOFISH_ALGORITHM_BYTE:
        return 32;
    default:
        return 0;
    }
}

void blockCfbInit(struct BlockCfbContext *context, byte const key[16], int Cipher)
{
    int k;
    context->Cipher = Cipher;
    context->blocksize = blockBlockSize(context->Cipher);
    context->keysize = blockKeySize(context->Cipher);
    switch (context->Cipher) {
    case IDEA_ALGORITHM_BYTE:
        ideaExpandKey(key, context->ctx.idea_key);
        k = 0; /* no selftest for IDEA */
        break;
    case TDES_ALGORITHM_BYTE:
        k = fsetkey3((byte *)key, context->ctx.ks);
        break;
    case CAST5_ALGORITHM_BYTE:
        k = cast_setkey(&context->ctx.C5ctx, (byte *)key, 16);
        break;
    case BLOWFISH_ALGORITHM_BYTE:
        k = Blowfish_Init(&context->ctx.bfctx, (byte *)key, 16);
        break;
    case TWOFISH_ALGORITHM_BYTE:
        k = makeKey(&context->ctx.tfctx, DIR_ENCRYPT, 256, (byte *)key);
        break;
    case AES128_ALGORITHM_BYTE:
        k = 0 >= rijndaelKeySetupEnc(context->ctx.rk, key, 128);
        break;
    case AES192_ALGORITHM_BYTE:
        k = 0 >= rijndaelKeySetupEnc(context->ctx.rk, key, 192);
        break;
    case AES256_ALGORITHM_BYTE:
        k = 0 >= rijndaelKeySetupEnc(context->ctx.rk, key, 256);
        break;
    default:
        break;
    }
    if (k)
        exitPGP(-1);
    blockCfbReinit(context, 0);
}

void blockCfbDestroy(struct BlockCfbContext *context)
{
    burn(*context);
}

/*
 * Okay, explanation time:
 * Phil invented a unique way of doing CFB that's sensitive to semantic
 * boundaries within the data being encrypted.  One way to phrase
 * CFB en/decryption is to say that you XOR the current 8 (blocksize) bytes with
 * IDEA/CAST5(previous 8 bytes of ciphertext).  Normally, you repeat this
 * at 8-byte intervals, but Phil decided to resync things on the
 * boundaries between elements in the stream being encrypted.
 *
 * That is, the last 4 bytes of a 12-byte field are en/decrypted using
 * the first 4 bytes of IDEA/CAST5(previous 8 bytes of ciphertext), but then
 * the last 4 bytes of that IDEA/CAST5 computation are thrown away, and the
 * first 8 bytes of the next field are en/decrypted using
 * IDEA/CAST5(last 8 bytes of ciphertext).  This is equivalent to using a
 * shorter feedback length (if you're familiar with the general CFB
 * technique) briefly, and doesn't weaken the cipher any (using shorter
 * CFB lengths makes it stronger, actually), it just makes it a bit unusual.
 *
 * Anyway, to accomodate this behaviour, every time we do an IDEA/CAST5
 * encrpytion of 8 bytes of ciphertext to get 8 bytes of XOR mask,
 * we remember the ciphertext.  Then if we have to resync things
 * after having processed, say, 2 bytes, we refill the iv buffer
 * with the last 6 bytes of the old ciphertext followed by the
 * 2 bytes of new ciphertext stored in the front of the iv buffer.
 */
void blockCfbSync(struct BlockCfbContext *context)
{
    int bufleft = context->bufleft;

    if (bufleft) {
	    memmove(context->iv + bufleft, context->iv, context->blocksize - bufleft);
	    memcpy(context->iv, context->oldcipher + context->blocksize - bufleft, bufleft);
	    context->bufleft = 0;
    }
}

/*
 * Encrypt a buffer of data, using IDEA/CAST5 in CFB mode.
 * There are more compact ways of writing this, but this is
 * written for speed.
 */
void blockCfbEncrypt(struct BlockCfbContext *context, byte const *src,
		    byte * dest, int count)
{
    int bufleft = context->bufleft;
    byte *bufptr = context->iv + context->blocksize - bufleft;

    /* If there are no more bytes to encrypt that there are bytes
     * in the buffer, XOR them in and return.
     */
    if (count <= bufleft) {
	    context->bufleft = bufleft - count;
	    while (count--) {
	        *dest++ = *bufptr++ ^= *src++;
	    }
	    return;
    }
    count -= bufleft;
    /* Encrypt the first bufleft (0 to 7) bytes of the input by XOR
     * with the last bufleft bytes in the iv buffer.
     */
    while (bufleft--) {
	    *dest++ = (*bufptr++ ^= *src++);
    }
    /* Encrypt middle blocks of the input by cranking the cipher,
     * XORing 8-byte blocks, and repeating until the count
     * is 8 or less.
     */
    while (count > context->blocksize) {
	    bufptr = context->iv;
	    memcpy(context->oldcipher, bufptr, context->blocksize);
	    switch (context->Cipher) {
	    case IDEA_ALGORITHM_BYTE:
		    ideaCipher(bufptr, bufptr, context->ctx.idea_key);
            break;
	    case CAST5_ALGORITHM_BYTE:
		    cast_encrypt_block(&context->ctx.C5ctx, bufptr, bufptr);
            break;
        case TDES_ALGORITHM_BYTE:
            fencrypt3(bufptr, 0, context->ctx.ks);
            break;
	    case BLOWFISH_ALGORITHM_BYTE:
	        Blowfish_Encrypt1(&context->ctx.bfctx, bufptr, bufptr);
            break;
        case TWOFISH_ALGORITHM_BYTE:
            blockEncrypt(&context->ctx.tfctx, bufptr, bufptr);
            break;
	    case AES128_ALGORITHM_BYTE:
            rijndaelEncrypt(context->ctx.rk, R_ROUNDS128, bufptr, bufptr);
            break;
	    case AES192_ALGORITHM_BYTE:
            rijndaelEncrypt(context->ctx.rk, R_ROUNDS192, bufptr, bufptr);
            break;
	    case AES256_ALGORITHM_BYTE:
            rijndaelEncrypt(context->ctx.rk, R_ROUNDS256, bufptr, bufptr);
            break;
	    default:
	        fill0(bufptr, context->blocksize);
	    }
	    bufleft = context->blocksize;
	    count -= context->blocksize;
	    do {
	        *dest++ = (*bufptr++ ^= *src++);
	    } while (--bufleft);
    }
    /* Do the last 1 to 8 bytes */
    bufptr = context->iv;
    memcpy(context->oldcipher, bufptr, context->blocksize);
	switch (context->Cipher) {
	case IDEA_ALGORITHM_BYTE:
		ideaCipher(bufptr, bufptr, context->ctx.idea_key);
        break;
	case CAST5_ALGORITHM_BYTE:
		cast_encrypt_block(&context->ctx.C5ctx, bufptr, bufptr);
        break;
    case TDES_ALGORITHM_BYTE:
        fencrypt3(bufptr, 0, context->ctx.ks);
        break;
	case BLOWFISH_ALGORITHM_BYTE:
	    Blowfish_Encrypt1(&context->ctx.bfctx, bufptr, bufptr);
        break;
    case TWOFISH_ALGORITHM_BYTE:
        blockEncrypt(&context->ctx.tfctx, bufptr, bufptr);
        break;
	case AES128_ALGORITHM_BYTE:
        rijndaelEncrypt(context->ctx.rk, R_ROUNDS128, bufptr, bufptr);
        break;
	case AES192_ALGORITHM_BYTE:
        rijndaelEncrypt(context->ctx.rk, R_ROUNDS192, bufptr, bufptr);
        break;
	case AES256_ALGORITHM_BYTE:
        rijndaelEncrypt(context->ctx.rk, R_ROUNDS256, bufptr, bufptr);
        break;
	default:
	    fill0(bufptr, context->blocksize);
	}
    context->bufleft = context->blocksize - count;
    do {
	    *dest++ = (*bufptr++ ^= *src++);
    } while (--count);
}


/*
 * Decrypt a buffer of data, using IDEA/CAST5 in CFB mode.
 * There are more compact ways of writing this, but this is
 * written for speed.
 */
void blockCfbDecrypt(struct BlockCfbContext *context, byte const *src,
		    byte * dest, int count)
{
    int bufleft = context->bufleft;
    static byte *bufptr;
    byte t;

    bufptr = context->iv + (context->blocksize - bufleft);
    if (count <= bufleft) {
	    context->bufleft = bufleft - count;
	    while (count--) {
	        t = *bufptr;
	        *dest++ = t ^ (*bufptr++ = *src++);
	    }
	    return;
    }
    count -= bufleft;
    while (bufleft--) {
	    t = *bufptr;
	    *dest++ = t ^ (*bufptr++ = *src++);
    }
    while (count > context->blocksize) {
	    bufptr = context->iv;
	    memcpy(context->oldcipher, bufptr, context->blocksize);
	    switch (context->Cipher) {
	    case IDEA_ALGORITHM_BYTE:
		    ideaCipher(bufptr, bufptr, context->ctx.idea_key);
            break;
	    case CAST5_ALGORITHM_BYTE:
		    cast_encrypt_block(&context->ctx.C5ctx, bufptr, bufptr);
            break;
        case TDES_ALGORITHM_BYTE:
            fencrypt3(bufptr, 0, context->ctx.ks);
            break;
	    case BLOWFISH_ALGORITHM_BYTE:
	        Blowfish_Encrypt1(&context->ctx.bfctx, bufptr, bufptr);
            break;
        case TWOFISH_ALGORITHM_BYTE:
            blockEncrypt(&context->ctx.tfctx, bufptr, bufptr);
            break;
	    case AES128_ALGORITHM_BYTE:
            rijndaelEncrypt(context->ctx.rk, R_ROUNDS128, bufptr, bufptr);
            break;
	    case AES192_ALGORITHM_BYTE:
            rijndaelEncrypt(context->ctx.rk, R_ROUNDS192, bufptr, bufptr);
            break;
	    case AES256_ALGORITHM_BYTE:
            rijndaelEncrypt(context->ctx.rk, R_ROUNDS256, bufptr, bufptr);
            break;
	    default:
	        fill0(bufptr, context->blocksize);
	    }
	    bufleft = context->blocksize;
	    count -= context->blocksize;
	    do {
	        t = *bufptr;
	        *dest++ = t ^ (*bufptr++ = *src++);
	    } while (--bufleft);
    }
    bufptr = context->iv;
    memcpy(context->oldcipher, bufptr, context->blocksize);
	switch (context->Cipher) {
	case IDEA_ALGORITHM_BYTE:
		ideaCipher(bufptr, bufptr, context->ctx.idea_key);
        break;
	case CAST5_ALGORITHM_BYTE:
		cast_encrypt_block(&context->ctx.C5ctx, bufptr, bufptr);
        break;
    case TDES_ALGORITHM_BYTE:
        fencrypt3(bufptr, 0, context->ctx.ks);
        break;
	case BLOWFISH_ALGORITHM_BYTE:
	    Blowfish_Encrypt1(&context->ctx.bfctx, bufptr, bufptr);
        break;
    case TWOFISH_ALGORITHM_BYTE:
        blockEncrypt(&context->ctx.tfctx, bufptr, bufptr);
        break;
	case AES128_ALGORITHM_BYTE:
        rijndaelEncrypt(context->ctx.rk, R_ROUNDS128, bufptr, bufptr);
        break;
	case AES192_ALGORITHM_BYTE:
        rijndaelEncrypt(context->ctx.rk, R_ROUNDS192, bufptr, bufptr);
        break;
	case AES256_ALGORITHM_BYTE:
        rijndaelEncrypt(context->ctx.rk, R_ROUNDS256, bufptr, bufptr);
        break;
	default:
	    fill0(bufptr, context->blocksize);
	}
    context->bufleft = context->blocksize - count;
    do {
	    t = *bufptr;
	    *dest++ = t ^ (*bufptr++ = *src++);
    } while (--count);
}

/********************************************************************/

/*
 * Cryptographically strong pseudo-random-number generator.
 * The design is from Appendix C of ANSI X9.17, "Financial
 * Institution Key Management (Wholesale)", with CAST5
 * substituted for the DES.
 */

/*
 * Initialize a cryptographic random-number generator.
 * key and seed should be arbitrary.
 */
void blockRandInit(struct BlockRandContext *context, byte const key[16],
		  byte const seed[8])
{
    /*int i;*/

    cast_setkey(&context->C5ctx, (byte *)key, 16);
    context->bufleft = 0;
    memcpy(context->internalbuf, seed, 8);
}


/*
 * Read out the RNG's state.
 */
void blockRandState(struct BlockRandContext *context, byte key[16], byte seed[8])
{
    int i;

    memcpy(seed, context->internalbuf, 8);
    for (i = 0; i < 8; i++) {
	    /*key[2 * i] = context->key[i] >> 8;*/
	    /*key[2 * i + 1] = context->key[i];*/
	    key[2 * i]     = (byte)(context->C5ctx.Km[i] >> 8);
	    key[2 * i + 1] = (byte)(context->C5ctx.Km[i]);
    }

}

/*
 * Encrypt the RNG's state with the given CFB encryptor.
 */
void blockRandWash(struct BlockRandContext *context, struct BlockCfbContext *cfb)
{
    byte keyseed[16 + 8];
    /*int i;*/

    blockRandState(context, keyseed, keyseed + 16);
    blockCfbEncrypt(cfb, keyseed, keyseed, 16 + 8);
    blockRandInit(context, keyseed, keyseed + 16);

    memset(keyseed, 0, 16 + 8);
}

/*
 * Cryptographic pseudo-random-number generator, used for generating
 * session keys.
 */
byte
blockRandByte(struct BlockRandContext *c)
{
    int i;

    if (!c->bufleft) {
    	byte timestamp[8];

	    /* Get some true-random noise to help */
	    randPoolGetBytes(timestamp, sizeof(timestamp));

	    /* Compute next 8 bytes of output */
	    for (i = 0; i < 8; i++)
	        c->outbuf[i] = c->internalbuf[i] ^ timestamp[i];
	    cast_encrypt_block(&c->C5ctx, c->outbuf, c->outbuf);
	    /* Compute new seed vector */
	    for (i = 0; i < 8; i++)
	        c->internalbuf[i] = c->outbuf[i] ^ timestamp[i];
	    cast_encrypt_block(&c->C5ctx, c->internalbuf, c->internalbuf);
	    burn(timestamp);
	    c->bufleft = 8;
    }
    return c->outbuf[--c->bufleft];
}
