
/*
 *  haval.c:  specifies the routines in the HAVAL (V.1) hashing library.
 *
 *      HAVAL is a one-way hashing algorithm with the following
 *      collision-resistant property:
 *             It is computationally infeasible to find two or more
 *             messages that are hashed into the same fingerprint.
 *      
 *  Reference:
 *       Y. Zheng, J. Pieprzyk and J. Seberry:
 *       ``HAVAL --- a one-way hashing algorithm with variable
 *       length of output'', Advances in Cryptology --- AUSCRYPT'92,
 *       Lecture Notes in Computer Science,  Vol.718, pp.83-104, 
 *       Springer-Verlag, 1993.
 *
 *  Descriptions:
 *      -  haval_string:      hash a string
 *      -  haval_file:        hash a file
 *      -  haval_stdin:       filter -- hash input from the stdin device
 *      -  haval_hash:        hash a string of specified length
 *                            (Haval_hash is used in conjunction with
 *                             haval_start & haval_end.)
 *      -  haval_hash_block:  hash a 32-word block
 *      -  haval_start:       initialization
 *      -  haval_end:         finalization
 *
 *  Author:     Yuliang Zheng
 *              School of Computing and Information Technology
 *              Monash University
 *              McMahons Road, Frankston 3199, Australia
 *              Email: yzheng@fcit.monash.edu.au
 *              URL:   http://www-pscit.fcit.monash.edu.au:/~yuliang/
 *              Voice: +61 3 9904 4196
 *              Fax  : +61 3 9904 4124
 *
 *  Date:
 *              First release: June  1993
 *
 *              Revised:       April 1996
 *                       Major changes:
 *                       1. added    #include <string.h>
 *                       2. deleted  #include <memory.h>
 *                       3. appended 'L' to all 32-bit constants
 *
 *                       Many thanks to Ross Williams (ross@rocksoft.com)
 *                       for his invaluable comments on the previous
 *                       version of the code.
 *
 *              Revised:       April 1997
 *                       Corrected the first byte in padding[128]
 *                       (from 0x80 to 0x01).
 *
 *                       Thanks go to Paulo Barreto (pbarreto@uninet.com.br)
 *                       who detected the padding error while implementing
 *                       HAVAL independently.
 *
 *              Revised:       April 2001 by Disastry
 *                       1. Number of passes and digest size can be changed
 *                       without recompiling
 *                       2. Haval rounds now done in for loop if HAVAL_LOOP
 *                       defined (code was so large that did not compile with
 *                       some compilers, now it does.. and speed is almost the same)
 *
 *      Copyright (C) 1997 by Yuliang Zheng.  All rights reserved. 
 *      This program may not be sold or used as inducement to sell
 *      a  product without the  written  permission of the author.
 */

#define HAVAL_LOOP

#include <stdio.h>
#include <string.h>
#include "haval.h"

#define VERSION    1                         /* current version number */

void haval_string (char *, unsigned char *); /* hash a string */
int  haval_file (char *, unsigned char *);   /* hash a file */
void haval_stdin (void);                     /* hash input from stdin */
void haval_start (haval_state *);            /* initialization */
void haval_hash (haval_state *,
        unsigned char *, unsigned int);      /* updating routine */
void haval_end (haval_state *, unsigned char *); /* finalization */
void haval_hash_block (haval_state *);       /* hash a 32-word block */
static void haval_tailor (haval_state *);    /* folding the last output */

static unsigned char padding[128] = {        /* constants for padding */
0x01, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

#define f_1(x6, x5, x4, x3, x2, x1, x0)          \
           ((x1) & ((x0) ^ (x4)) ^ (x2) & (x5) ^ \
            (x3) & (x6) ^ (x0))

#define f_2(x6, x5, x4, x3, x2, x1, x0)                         \
           ((x2) & ((x1) & ~(x3) ^ (x4) & (x5) ^ (x6) ^ (x0)) ^ \
            (x4) & ((x1) ^ (x5)) ^ (x3) & (x5) ^ (x0)) 

#define f_3(x6, x5, x4, x3, x2, x1, x0)          \
           ((x3) & ((x1) & (x2) ^ (x6) ^ (x0)) ^ \
            (x1) & (x4) ^ (x2) & (x5) ^ (x0))

#define f_4(x6, x5, x4, x3, x2, x1, x0)                                 \
           ((x4) & ((x5) & ~(x2) ^ (x3) & ~(x6) ^ (x1) ^ (x6) ^ (x0)) ^ \
            (x3) & ((x1) & (x2) ^ (x5) ^ (x6)) ^                        \
            (x2) & (x6) ^ (x0))

#define f_5(x6, x5, x4, x3, x2, x1, x0)             \
           ((x0) & ((x1) & (x2) & (x3) ^ ~(x5)) ^   \
            (x1) & (x4) ^ (x2) & (x5) ^ (x3) & (x6))

/*
 * Permutations phi_{i,j}, i=3,4,5, j=1,...,i.
 *
 * PASS = 3:
 *               6 5 4 3 2 1 0
 *               | | | | | | | (replaced by)
 *  phi_{3,1}:   1 0 3 5 6 2 4
 *  phi_{3,2}:   4 2 1 0 5 3 6
 *  phi_{3,3}:   6 1 2 3 4 5 0
 *
 * PASS = 4:
 *               6 5 4 3 2 1 0
 *               | | | | | | | (replaced by)
 *  phi_{4,1}:   2 6 1 4 5 3 0
 *  phi_{4,2}:   3 5 2 0 1 6 4
 *  phi_{4,3}:   1 4 3 6 0 2 5
 *  phi_{4,4}:   6 4 0 5 2 1 3
 *
 * PASS = 5:
 *               6 5 4 3 2 1 0
 *               | | | | | | | (replaced by)
 *  phi_{5,1}:   3 4 1 0 5 2 6
 *  phi_{5,2}:   6 2 1 0 3 4 5
 *  phi_{5,3}:   2 6 0 4 3 1 5
 *  phi_{5,4}:   1 5 3 2 0 4 6
 *  phi_{5,5}:   2 5 0 6 4 3 1
 */

#define Fphi_1_3(x6, x5, x4, x3, x2, x1, x0) \
           f_1(x1, x0, x3, x5, x6, x2, x4)
#define Fphi_1_4(x6, x5, x4, x3, x2, x1, x0) \
           f_1(x2, x6, x1, x4, x5, x3, x0)
#define Fphi_1_5(x6, x5, x4, x3, x2, x1, x0) \
           f_1(x3, x4, x1, x0, x5, x2, x6)

#define Fphi_2_3(x6, x5, x4, x3, x2, x1, x0) \
           f_2(x4, x2, x1, x0, x5, x3, x6)
#define Fphi_2_4(x6, x5, x4, x3, x2, x1, x0) \
           f_2(x3, x5, x2, x0, x1, x6, x4)
#define Fphi_2_5(x6, x5, x4, x3, x2, x1, x0) \
           f_2(x6, x2, x1, x0, x3, x4, x5)

#define Fphi_3_3(x6, x5, x4, x3, x2, x1, x0) \
           f_3(x6, x1, x2, x3, x4, x5, x0)
#define Fphi_3_4(x6, x5, x4, x3, x2, x1, x0) \
           f_3(x1, x4, x3, x6, x0, x2, x5)
#define Fphi_3_5(x6, x5, x4, x3, x2, x1, x0) \
           f_3(x2, x6, x0, x4, x3, x1, x5)

#define Fphi_4_4(x6, x5, x4, x3, x2, x1, x0) \
           f_4(x6, x4, x0, x5, x2, x1, x3)
#define Fphi_4_5(x6, x5, x4, x3, x2, x1, x0) \
            f_4(x1, x5, x3, x2, x0, x4, x6)

#define Fphi_5_5(x6, x5, x4, x3, x2, x1, x0) \
           f_5(x2, x5, x0, x6, x4, x3, x1)

#define rotate_right(x, n) (((x) >> (n)) | ((x) << (32-(n))))

#define FF_1_3(x7, x6, x5, x4, x3, x2, x1, x0, w) {                        \
      register haval_word temp = Fphi_1_3(x6, x5, x4, x3, x2, x1, x0);     \
      (x7) = rotate_right(temp, 7) + rotate_right((x7), 11) + (w);       \
      }

#define FF_2_3(x7, x6, x5, x4, x3, x2, x1, x0, w, c) {                      \
      register haval_word temp = Fphi_2_3(x6, x5, x4, x3, x2, x1, x0);      \
      (x7) = rotate_right(temp, 7) + rotate_right((x7), 11) + (w) + (c);  \
      }

#define FF_3_3(x7, x6, x5, x4, x3, x2, x1, x0, w, c) {                      \
      register haval_word temp = Fphi_3_3(x6, x5, x4, x3, x2, x1, x0);      \
      (x7) = rotate_right(temp, 7) + rotate_right((x7), 11) + (w) + (c);  \
      }

#define FF_1_4(x7, x6, x5, x4, x3, x2, x1, x0, w) {                        \
      register haval_word temp = Fphi_1_4(x6, x5, x4, x3, x2, x1, x0);     \
      (x7) = rotate_right(temp, 7) + rotate_right((x7), 11) + (w);       \
      }

#define FF_2_4(x7, x6, x5, x4, x3, x2, x1, x0, w, c) {                      \
      register haval_word temp = Fphi_2_4(x6, x5, x4, x3, x2, x1, x0);      \
      (x7) = rotate_right(temp, 7) + rotate_right((x7), 11) + (w) + (c);  \
      }

#define FF_3_4(x7, x6, x5, x4, x3, x2, x1, x0, w, c) {                      \
      register haval_word temp = Fphi_3_4(x6, x5, x4, x3, x2, x1, x0);      \
      (x7) = rotate_right(temp, 7) + rotate_right((x7), 11) + (w) + (c);  \
      }

#define FF_4_4(x7, x6, x5, x4, x3, x2, x1, x0, w, c) {                      \
      register haval_word temp = Fphi_4_4(x6, x5, x4, x3, x2, x1, x0);      \
      (x7) = rotate_right(temp, 7) + rotate_right((x7), 11) + (w) + (c);  \
      }

#define FF_1_5(x7, x6, x5, x4, x3, x2, x1, x0, w) {                        \
      register haval_word temp = Fphi_1_5(x6, x5, x4, x3, x2, x1, x0);     \
      (x7) = rotate_right(temp, 7) + rotate_right((x7), 11) + (w);       \
      }

#define FF_2_5(x7, x6, x5, x4, x3, x2, x1, x0, w, c) {                      \
      register haval_word temp = Fphi_2_5(x6, x5, x4, x3, x2, x1, x0);      \
      (x7) = rotate_right(temp, 7) + rotate_right((x7), 11) + (w) + (c);  \
      }

#define FF_3_5(x7, x6, x5, x4, x3, x2, x1, x0, w, c) {                      \
      register haval_word temp = Fphi_3_5(x6, x5, x4, x3, x2, x1, x0);      \
      (x7) = rotate_right(temp, 7) + rotate_right((x7), 11) + (w) + (c);  \
      }

#define FF_4_5(x7, x6, x5, x4, x3, x2, x1, x0, w, c) {                      \
      register haval_word temp = Fphi_4_5(x6, x5, x4, x3, x2, x1, x0);      \
      (x7) = rotate_right(temp, 7) + rotate_right((x7), 11) + (w) + (c);  \
      }

#define FF_5_5(x7, x6, x5, x4, x3, x2, x1, x0, w, c) {                      \
      register haval_word temp = Fphi_5_5(x6, x5, x4, x3, x2, x1, x0);      \
      (x7) = rotate_right(temp, 7) + rotate_right((x7), 11) + (w) + (c);  \
      }

/*
 * translate every four characters into a word.
 * assume the number of characters is a multiple of four.
 */
#define ch2uint(string, word, slen) {      \
  unsigned char *sp = string;              \
  haval_word    *wp = word;                \
  while (sp < (string) + (slen)) {         \
    *wp++ =  (haval_word)*sp            |  \
            ((haval_word)*(sp+1) <<  8) |  \
            ((haval_word)*(sp+2) << 16) |  \
            ((haval_word)*(sp+3) << 24);   \
    sp += 4;                               \
  }                                        \
}

/* translate each word into four characters */
#define uint2ch(word, string, wlen) {              \
  haval_word    *wp = word;                        \
  unsigned char *sp = string;                      \
  while (wp < (word) + (wlen)) {                   \
    *(sp++) = (unsigned char)( *wp        & 0xFF); \
    *(sp++) = (unsigned char)((*wp >>  8) & 0xFF); \
    *(sp++) = (unsigned char)((*wp >> 16) & 0xFF); \
    *(sp++) = (unsigned char)((*wp >> 24) & 0xFF); \
    wp++;                                          \
  }                                                \
}


/* hash a string */
void haval_string (char *string, unsigned char *fingerprint)
{
  haval_state   state;
  unsigned int  len = strlen (string);

  haval_start (&state);
  haval_hash (&state, (unsigned char *)string, len);
  haval_end (&state, fingerprint);
}

/* hash a file */
int haval_file (char *file_name, unsigned char *fingerprint)
{
  FILE          *file;
  haval_state   state;
  int           len;
  unsigned char buffer[1024];

  if ((file = fopen (file_name, "rb")) == NULL){
    return (1);                                    /* fail */
  } else {
    haval_start (&state);
    while (len = fread (buffer, 1, 1024, file)) {
      haval_hash (&state, buffer, len);
    }
    fclose (file);
    haval_end (&state, fingerprint);
    return (0);                                    /* success */
 }
}

/* hash input from stdin */
void haval_stdin (void)
{
  haval_state   state;
  int           i, len;
  unsigned char buffer[32],
                fingerprint[32];

  haval_start (&state);
  while (len = fread (buffer, 1, 32, stdin)) {
    haval_hash (&state, buffer, len);
  }
  haval_end (&state, fingerprint);
  
  for (i = 0; i < state.fptlen >> 3; i++) {
    putchar(fingerprint[i]);
  }
}

/* initialization */
void haval_start (haval_state *state)
{
    state->count[0]       = state->count[1] = 0;   /* clear count */
    state->fingerprint[0] = 0x243F6A88L;           /* initial fingerprint */
    state->fingerprint[1] = 0x85A308D3L;
    state->fingerprint[2] = 0x13198A2EL;
    state->fingerprint[3] = 0x03707344L;
    state->fingerprint[4] = 0xA4093822L;
    state->fingerprint[5] = 0x299F31D0L;
    state->fingerprint[6] = 0x082EFA98L;
    state->fingerprint[7] = 0xEC4E6C89L;
/* if the pass and fptlen is not defined or is wrong set 3 and 128 */
if (state->pass<3 || state->pass>5) state->pass = 3;
if (state->fptlen<128 || state->fptlen>256 || state->fptlen%32) state->fptlen = 128;
}

/*
 * hash a string of specified length.
 * to be used in conjunction with haval_start and haval_end.
 */
void haval_hash (haval_state *state,
                 unsigned char *str, unsigned int str_len)
{
  unsigned int i,
           rmd_len,
           fill_len;

  /* calculate the number of bytes in the remainder */
  rmd_len  = (unsigned int)((state->count[0] >> 3) & 0x7F);
  fill_len = 128 - rmd_len;

  /* update the number of bits */
  if ((state->count[0] +=  (haval_word)str_len << 3)
                        < ((haval_word)str_len << 3)) {
     state->count[1]++;
  }
  state->count[1] += (haval_word)str_len >> 29;

#ifdef LITTLE_ENDIAN

  /* hash as many blocks as possible */
  if (rmd_len + str_len >= 128) {
    memcpy (((unsigned char *)state->block)+rmd_len, str, fill_len);
    haval_hash_block (state);
    for (i = fill_len; i + 127 < str_len; i += 128){
      memcpy ((unsigned char *)state->block, str+i, 128);
      haval_hash_block (state);
    }
    rmd_len = 0;
  } else {
    i = 0;
  }
  memcpy (((unsigned char *)state->block)+rmd_len, str+i, str_len-i);

#else

  /* hash as many blocks as possible */
  if (rmd_len + str_len >= 128) {
    memcpy (&state->remainder[rmd_len], str, fill_len);
    ch2uint(state->remainder, state->block, 128);
    haval_hash_block (state);
    for (i = fill_len; i + 127 < str_len; i += 128){
      memcpy (state->remainder, str+i, 128);
      ch2uint(state->remainder, state->block, 128);
      haval_hash_block (state);
    }
    rmd_len = 0;
  } else {
    i = 0;
  }
  /* save the remaining input chars */
  memcpy (&state->remainder[rmd_len], str+i, str_len-i);

#endif
}

/* finalization */
void haval_end (haval_state *state, unsigned char *final_fpt)
{
  unsigned char tail[10];
  unsigned int  rmd_len, pad_len;

  /*
   * save the version number, the number of passes, the fingerprint 
   * length and the number of bits in the unpadded message.
   */
  tail[0] = (unsigned char)(((state->fptlen  & 0x3) << 6) |
                            ((state->pass    & 0x7) << 3) |
                             (VERSION & 0x7));
  tail[1] = (unsigned char)((state->fptlen >> 2) & 0xFF);
  uint2ch (state->count, &tail[2], 2);

  /* pad out to 118 mod 128 */
  rmd_len = (unsigned int)((state->count[0] >> 3) & 0x7f);
  pad_len = (rmd_len < 118) ? (118 - rmd_len) : (246 - rmd_len);
  haval_hash (state, padding, pad_len);

  /*
   * append the version number, the number of passes,
   * the fingerprint length and the number of bits
   */
  haval_hash (state, tail, 10);

  /* tailor the last output */
  haval_tailor(state);

  /* translate and save the final fingerprint */
  uint2ch (state->fingerprint, final_fpt, state->fptlen >> 5);

  /* clear the state information */
  memset ((unsigned char *)state, 0, sizeof (*state));
}

#ifdef HAVAL_LOOP
unsigned p2a[32] = { 5,14,26,18,11,28, 7,16,
                     0,23,20,22, 1,10, 4, 8,
                    30, 3,21, 9,17,24,29, 6,
                    19,12,15,13, 2,25,31,27};
haval_word p2b[32] = {  0x452821E6L, 0x38D01377L, 0xBE5466CFL, 0x34E90C6CL,
                        0xC0AC29B7L, 0xC97C50DDL, 0x3F84D5B5L, 0xB5470917L,
                        0x9216D5D9L, 0x8979FB1BL, 0xD1310BA6L, 0x98DFB5ACL,
                        0x2FFD72DBL, 0xD01ADFB7L, 0xB8E1AFEDL, 0x6A267E96L,
                        0xBA7C9045L, 0xF12C7F99L, 0x24A19947L, 0xB3916CF7L,
                        0x0801F2E2L, 0x858EFC16L, 0x636920D8L, 0x71574E69L,
                        0xA458FEA3L, 0xF4933D7EL, 0x0D95748FL, 0x728EB658L,
                        0x718BCD58L, 0x82154AEEL, 0x7B54A41DL, 0xC25A59B5L};

unsigned p3a[32] = {19, 9, 4,20,28,17, 8,22,
                    29,14,25,12,24,30,16,26,
                    31,15, 7, 3, 1, 0,18,27,
                    13, 6,21,10,23,11, 5, 2};
haval_word p3b[32] = {  0x9C30D539L, 0x2AF26013L, 0xC5D1B023L, 0x286085F0L,
                        0xCA417918L, 0xB8DB38EFL, 0x8E79DCB0L, 0x603A180EL,
                        0x6C9E0E8BL, 0xB01E8A3EL, 0xD71577C1L, 0xBD314B27L,
                        0x78AF2FDAL, 0x55605C60L, 0xE65525F3L, 0xAA55AB94L,
                        0x57489862L, 0x63E81440L, 0x55CA396AL, 0x2AAB10B6L,
                        0xB4CC5C34L, 0x1141E8CEL, 0xA15486AFL, 0x7C72E993L,
                        0xB3EE1411L, 0x636FBC2AL, 0x2BA9C55DL, 0x741831F6L,
                        0xCE5C3E16L, 0x9B87931EL, 0xAFD6BA33L, 0x6C24CF5CL};

unsigned p4a[32] = {24, 4, 0,14, 2, 7,28,23,
                    26, 6,30,20,18,25,19, 3,
                    22,11,31,21, 8,27,12, 9,
                     1,29, 5,15,17,10,16,13};
haval_word p4b[32] = {  0x7A325381L, 0x28958677L, 0x3B8F4898L, 0x6B4BB9AFL,
                        0xC4BFE81BL, 0x66282193L, 0x61D809CCL, 0xFB21A991L,
                        0x487CAC60L, 0x5DEC8032L, 0xEF845D5DL, 0xE98575B1L,
                        0xDC262302L, 0xEB651B88L, 0x23893E81L, 0xD396ACC5L,
                        0x0F6D6FF3L, 0x83F44239L, 0x2E0B4482L, 0xA4842004L,
                        0x69C8F04AL, 0x9E1F9B5EL, 0x21C66842L, 0xF6E96C9AL,
                        0x670C9C61L, 0xABD388F0L, 0x6A51A0D2L, 0xD8542F68L,
                        0x960FA728L, 0xAB5133A3L, 0x6EEF0B6CL, 0x137A3BE4L};

unsigned p5a[32] = {27, 3,21,26,17,11,20,29,
                    19, 0,12, 7,13, 8,31,10,
                     5, 9,14,30,18, 6,28,24,
                     2,23,16,22, 4, 1,25,15};
haval_word p5b[32] = {  0xBA3BF050L, 0x7EFB2A98L, 0xA1F1651DL, 0x39AF0176L, 
                        0x66CA593EL, 0x82430E88L, 0x8CEE8619L, 0x456F9FB4L, 
                        0x7D84A5C3L, 0x3B8B5EBEL, 0xE06F75D8L, 0x85C12073L, 
                        0x401A449FL, 0x56C16AA6L, 0x4ED3AA62L, 0x363F7706L, 
                        0x1BFEDF72L, 0x429B023DL, 0x37D0D724L, 0xD00A1248L, 
                        0xDB0FEAD3L, 0x49F1C09BL, 0x075372C9L, 0x80991B7BL, 
                        0x25D479D8L, 0xF6E8DEF7L, 0xE3FE501AL, 0xB6794C3BL, 
                        0x976CE0BDL, 0x04C006BAL, 0xC1A94FB6L, 0x409F60C4L};
#endif /* HAVAL_LOOP */

/* hash a 32-word block */
void haval_hash_block (haval_state *state)
{
  register haval_word t0 = state->fingerprint[0],    /* make use of */
                      t1 = state->fingerprint[1],    /* internal registers */
                      t2 = state->fingerprint[2],
                      t3 = state->fingerprint[3],
                      t4 = state->fingerprint[4],
                      t5 = state->fingerprint[5],
                      t6 = state->fingerprint[6],
                      t7 = state->fingerprint[7],
                      *w = state->block;

#ifndef HAVAL_LOOP

if (state->pass == 3) {
  /* Pass 1 */
  FF_1_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w   ));
  FF_1_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 1));
  FF_1_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ 2));
  FF_1_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ 3));
  FF_1_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 4));
  FF_1_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ 5));
  FF_1_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 6));
  FF_1_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 7));

  FF_1_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ 8));
  FF_1_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 9));
  FF_1_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+10));
  FF_1_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+11));
  FF_1_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+12));
  FF_1_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w+13));
  FF_1_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+14));
  FF_1_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+15));

  FF_1_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w+16));
  FF_1_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+17));
  FF_1_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+18));
  FF_1_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+19));
  FF_1_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+20));
  FF_1_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w+21));
  FF_1_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+22));
  FF_1_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+23));

  FF_1_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w+24));
  FF_1_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+25));
  FF_1_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+26));
  FF_1_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+27));
  FF_1_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+28));
  FF_1_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w+29));
  FF_1_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+30));
  FF_1_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+31));

  /* Pass 2 */
  FF_2_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ 5), 0x452821E6L);
  FF_2_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+14), 0x38D01377L);
  FF_2_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+26), 0xBE5466CFL);
  FF_2_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+18), 0x34E90C6CL);
  FF_2_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+11), 0xC0AC29B7L);
  FF_2_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w+28), 0xC97C50DDL);
  FF_2_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 7), 0x3F84D5B5L);
  FF_2_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+16), 0xB5470917L);

  FF_2_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w   ), 0x9216D5D9L);
  FF_2_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+23), 0x8979FB1BL);
  FF_2_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+20), 0xD1310BA6L);
  FF_2_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+22), 0x98DFB5ACL);
  FF_2_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 1), 0x2FFD72DBL);
  FF_2_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w+10), 0xD01ADFB7L);
  FF_2_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 4), 0xB8E1AFEDL);
  FF_2_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 8), 0x6A267E96L);

  FF_2_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w+30), 0xBA7C9045L);
  FF_2_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 3), 0xF12C7F99L);
  FF_2_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+21), 0x24A19947L);
  FF_2_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ 9), 0xB3916CF7L);
  FF_2_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+17), 0x0801F2E2L);
  FF_2_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w+24), 0x858EFC16L);
  FF_2_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+29), 0x636920D8L);
  FF_2_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 6), 0x71574E69L);

  FF_2_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w+19), 0xA458FEA3L);
  FF_2_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+12), 0xF4933D7EL);
  FF_2_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+15), 0x0D95748FL);
  FF_2_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+13), 0x728EB658L);
  FF_2_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 2), 0x718BCD58L);
  FF_2_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w+25), 0x82154AEEL);
  FF_2_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+31), 0x7B54A41DL);
  FF_2_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+27), 0xC25A59B5L);

  /* Pass 3 */
  FF_3_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w+19), 0x9C30D539L);
  FF_3_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 9), 0x2AF26013L);
  FF_3_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ 4), 0xC5D1B023L);
  FF_3_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+20), 0x286085F0L);
  FF_3_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+28), 0xCA417918L);
  FF_3_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w+17), 0xB8DB38EFL);
  FF_3_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 8), 0x8E79DCB0L);
  FF_3_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+22), 0x603A180EL);

  FF_3_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w+29), 0x6C9E0E8BL);
  FF_3_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+14), 0xB01E8A3EL);
  FF_3_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+25), 0xD71577C1L);
  FF_3_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+12), 0xBD314B27L);
  FF_3_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+24), 0x78AF2FDAL);
  FF_3_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w+30), 0x55605C60L);
  FF_3_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+16), 0xE65525F3L);
  FF_3_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+26), 0xAA55AB94L);

  FF_3_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w+31), 0x57489862L);
  FF_3_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+15), 0x63E81440L);
  FF_3_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ 7), 0x55CA396AL);
  FF_3_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ 3), 0x2AAB10B6L);
  FF_3_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 1), 0xB4CC5C34L);
  FF_3_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w   ), 0x1141E8CEL);
  FF_3_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+18), 0xA15486AFL);
  FF_3_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+27), 0x7C72E993L);

  FF_3_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w+13), 0xB3EE1411L);
  FF_3_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 6), 0x636FBC2AL);
  FF_3_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+21), 0x2BA9C55DL);
  FF_3_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+10), 0x741831F6L);
  FF_3_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+23), 0xCE5C3E16L);
  FF_3_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w+11), 0x9B87931EL);
  FF_3_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 5), 0xAFD6BA33L);
  FF_3_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 2), 0x6C24CF5CL);
}

if (state->pass == 4) {
  /* Pass 1 */
  FF_1_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w   ));
  FF_1_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 1));
  FF_1_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ 2));
  FF_1_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ 3));
  FF_1_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 4));
  FF_1_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ 5));
  FF_1_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 6));
  FF_1_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 7));

  FF_1_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ 8));
  FF_1_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 9));
  FF_1_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+10));
  FF_1_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+11));
  FF_1_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+12));
  FF_1_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+13));
  FF_1_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+14));
  FF_1_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+15));

  FF_1_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+16));
  FF_1_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+17));
  FF_1_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+18));
  FF_1_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+19));
  FF_1_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+20));
  FF_1_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+21));
  FF_1_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+22));
  FF_1_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+23));

  FF_1_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+24));
  FF_1_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+25));
  FF_1_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+26));
  FF_1_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+27));
  FF_1_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+28));
  FF_1_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+29));
  FF_1_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+30));
  FF_1_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+31));

  /* Pass 2 */
  FF_2_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ 5), 0x452821E6L);
  FF_2_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+14), 0x38D01377L);
  FF_2_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+26), 0xBE5466CFL);
  FF_2_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+18), 0x34E90C6CL);
  FF_2_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+11), 0xC0AC29B7L);
  FF_2_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+28), 0xC97C50DDL);
  FF_2_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 7), 0x3F84D5B5L);
  FF_2_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+16), 0xB5470917L);

  FF_2_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w   ), 0x9216D5D9L);
  FF_2_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+23), 0x8979FB1BL);
  FF_2_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+20), 0xD1310BA6L);
  FF_2_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+22), 0x98DFB5ACL);
  FF_2_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 1), 0x2FFD72DBL);
  FF_2_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+10), 0xD01ADFB7L);
  FF_2_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 4), 0xB8E1AFEDL);
  FF_2_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 8), 0x6A267E96L);

  FF_2_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+30), 0xBA7C9045L);
  FF_2_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 3), 0xF12C7F99L);
  FF_2_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+21), 0x24A19947L);
  FF_2_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ 9), 0xB3916CF7L);
  FF_2_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+17), 0x0801F2E2L);
  FF_2_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+24), 0x858EFC16L);
  FF_2_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+29), 0x636920D8L);
  FF_2_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 6), 0x71574E69L);

  FF_2_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+19), 0xA458FEA3L);
  FF_2_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+12), 0xF4933D7EL);
  FF_2_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+15), 0x0D95748FL);
  FF_2_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+13), 0x728EB658L);
  FF_2_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 2), 0x718BCD58L);
  FF_2_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+25), 0x82154AEEL);
  FF_2_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+31), 0x7B54A41DL);
  FF_2_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+27), 0xC25A59B5L);

  /* Pass 3 */
  FF_3_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+19), 0x9C30D539L);
  FF_3_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 9), 0x2AF26013L);
  FF_3_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ 4), 0xC5D1B023L);
  FF_3_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+20), 0x286085F0L);
  FF_3_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+28), 0xCA417918L);
  FF_3_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+17), 0xB8DB38EFL);
  FF_3_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 8), 0x8E79DCB0L);
  FF_3_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+22), 0x603A180EL);

  FF_3_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+29), 0x6C9E0E8BL);
  FF_3_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+14), 0xB01E8A3EL);
  FF_3_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+25), 0xD71577C1L);
  FF_3_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+12), 0xBD314B27L);
  FF_3_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+24), 0x78AF2FDAL);
  FF_3_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+30), 0x55605C60L);
  FF_3_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+16), 0xE65525F3L);
  FF_3_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+26), 0xAA55AB94L);

  FF_3_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+31), 0x57489862L);
  FF_3_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+15), 0x63E81440L);
  FF_3_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ 7), 0x55CA396AL);
  FF_3_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ 3), 0x2AAB10B6L);
  FF_3_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 1), 0xB4CC5C34L);
  FF_3_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w   ), 0x1141E8CEL);
  FF_3_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+18), 0xA15486AFL);
  FF_3_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+27), 0x7C72E993L);

  FF_3_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+13), 0xB3EE1411L);
  FF_3_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 6), 0x636FBC2AL);
  FF_3_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+21), 0x2BA9C55DL);
  FF_3_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+10), 0x741831F6L);
  FF_3_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+23), 0xCE5C3E16L);
  FF_3_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+11), 0x9B87931EL);
  FF_3_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 5), 0xAFD6BA33L);
  FF_3_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 2), 0x6C24CF5CL);

  /* Pass 4. executed only when PASS =4 or 5 */
  FF_4_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+24), 0x7A325381L);
  FF_4_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 4), 0x28958677L);
  FF_4_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w   ), 0x3B8F4898L);
  FF_4_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+14), 0x6B4BB9AFL);
  FF_4_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 2), 0xC4BFE81BL);
  FF_4_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ 7), 0x66282193L);
  FF_4_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+28), 0x61D809CCL);
  FF_4_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+23), 0xFB21A991L);

  FF_4_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+26), 0x487CAC60L);
  FF_4_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 6), 0x5DEC8032L);
  FF_4_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+30), 0xEF845D5DL);
  FF_4_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+20), 0xE98575B1L);
  FF_4_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+18), 0xDC262302L);
  FF_4_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+25), 0xEB651B88L);
  FF_4_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+19), 0x23893E81L);
  FF_4_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 3), 0xD396ACC5L);

  FF_4_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+22), 0x0F6D6FF3L);
  FF_4_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+11), 0x83F44239L);
  FF_4_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+31), 0x2E0B4482L);
  FF_4_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+21), 0xA4842004L);
  FF_4_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 8), 0x69C8F04AL);
  FF_4_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+27), 0x9E1F9B5EL);
  FF_4_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+12), 0x21C66842L);
  FF_4_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 9), 0xF6E96C9AL);

  FF_4_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ 1), 0x670C9C61L);
  FF_4_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+29), 0xABD388F0L);
  FF_4_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ 5), 0x6A51A0D2L);
  FF_4_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+15), 0xD8542F68L);
  FF_4_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+17), 0x960FA728L);
  FF_4_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+10), 0xAB5133A3L);
  FF_4_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+16), 0x6EEF0B6CL);
  FF_4_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+13), 0x137A3BE4L);
}

if (state->pass == 5) {
  /* Pass 1 */
  FF_1_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w   ));
  FF_1_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 1));
  FF_1_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ 2));
  FF_1_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ 3));
  FF_1_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 4));
  FF_1_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ 5));
  FF_1_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 6));
  FF_1_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 7));

  FF_1_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ 8));
  FF_1_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 9));
  FF_1_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+10));
  FF_1_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+11));
  FF_1_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+12));
  FF_1_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+13));
  FF_1_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+14));
  FF_1_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+15));

  FF_1_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+16));
  FF_1_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+17));
  FF_1_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+18));
  FF_1_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+19));
  FF_1_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+20));
  FF_1_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+21));
  FF_1_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+22));
  FF_1_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+23));

  FF_1_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+24));
  FF_1_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+25));
  FF_1_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+26));
  FF_1_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+27));
  FF_1_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+28));
  FF_1_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+29));
  FF_1_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+30));
  FF_1_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+31));

  /* Pass 2 */
  FF_2_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ 5), 0x452821E6L);
  FF_2_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+14), 0x38D01377L);
  FF_2_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+26), 0xBE5466CFL);
  FF_2_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+18), 0x34E90C6CL);
  FF_2_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+11), 0xC0AC29B7L);
  FF_2_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+28), 0xC97C50DDL);
  FF_2_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 7), 0x3F84D5B5L);
  FF_2_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+16), 0xB5470917L);

  FF_2_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w   ), 0x9216D5D9L);
  FF_2_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+23), 0x8979FB1BL);
  FF_2_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+20), 0xD1310BA6L);
  FF_2_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+22), 0x98DFB5ACL);
  FF_2_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 1), 0x2FFD72DBL);
  FF_2_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+10), 0xD01ADFB7L);
  FF_2_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 4), 0xB8E1AFEDL);
  FF_2_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 8), 0x6A267E96L);

  FF_2_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+30), 0xBA7C9045L);
  FF_2_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 3), 0xF12C7F99L);
  FF_2_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+21), 0x24A19947L);
  FF_2_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ 9), 0xB3916CF7L);
  FF_2_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+17), 0x0801F2E2L);
  FF_2_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+24), 0x858EFC16L);
  FF_2_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+29), 0x636920D8L);
  FF_2_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 6), 0x71574E69L);

  FF_2_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+19), 0xA458FEA3L);
  FF_2_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+12), 0xF4933D7EL);
  FF_2_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+15), 0x0D95748FL);
  FF_2_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+13), 0x728EB658L);
  FF_2_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 2), 0x718BCD58L);
  FF_2_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+25), 0x82154AEEL);
  FF_2_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+31), 0x7B54A41DL);
  FF_2_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+27), 0xC25A59B5L);

  /* Pass 3 */
  FF_3_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+19), 0x9C30D539L);
  FF_3_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 9), 0x2AF26013L);
  FF_3_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ 4), 0xC5D1B023L);
  FF_3_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+20), 0x286085F0L);
  FF_3_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+28), 0xCA417918L);
  FF_3_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+17), 0xB8DB38EFL);
  FF_3_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 8), 0x8E79DCB0L);
  FF_3_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+22), 0x603A180EL);

  FF_3_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+29), 0x6C9E0E8BL);
  FF_3_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+14), 0xB01E8A3EL);
  FF_3_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+25), 0xD71577C1L);
  FF_3_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+12), 0xBD314B27L);
  FF_3_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+24), 0x78AF2FDAL);
  FF_3_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+30), 0x55605C60L);
  FF_3_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+16), 0xE65525F3L);
  FF_3_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+26), 0xAA55AB94L);

  FF_3_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+31), 0x57489862L);
  FF_3_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+15), 0x63E81440L);
  FF_3_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ 7), 0x55CA396AL);
  FF_3_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ 3), 0x2AAB10B6L);
  FF_3_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 1), 0xB4CC5C34L);
  FF_3_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w   ), 0x1141E8CEL);
  FF_3_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+18), 0xA15486AFL);
  FF_3_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+27), 0x7C72E993L);

  FF_3_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+13), 0xB3EE1411L);
  FF_3_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 6), 0x636FBC2AL);
  FF_3_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+21), 0x2BA9C55DL);
  FF_3_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+10), 0x741831F6L);
  FF_3_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+23), 0xCE5C3E16L);
  FF_3_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+11), 0x9B87931EL);
  FF_3_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ 5), 0xAFD6BA33L);
  FF_3_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 2), 0x6C24CF5CL);

  /* Pass 4. executed only when PASS =4 or 5 */
  FF_4_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+24), 0x7A325381L);
  FF_4_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 4), 0x28958677L);
  FF_4_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w   ), 0x3B8F4898L);
  FF_4_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+14), 0x6B4BB9AFL);
  FF_4_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 2), 0xC4BFE81BL);
  FF_4_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ 7), 0x66282193L);
  FF_4_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+28), 0x61D809CCL);
  FF_4_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+23), 0xFB21A991L);

  FF_4_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+26), 0x487CAC60L);
  FF_4_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 6), 0x5DEC8032L);
  FF_4_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+30), 0xEF845D5DL);
  FF_4_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+20), 0xE98575B1L);
  FF_4_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+18), 0xDC262302L);
  FF_4_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+25), 0xEB651B88L);
  FF_4_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+19), 0x23893E81L);
  FF_4_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 3), 0xD396ACC5L);

  FF_4_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+22), 0x0F6D6FF3L);
  FF_4_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+11), 0x83F44239L);
  FF_4_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+31), 0x2E0B4482L);
  FF_4_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+21), 0xA4842004L);
  FF_4_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 8), 0x69C8F04AL);
  FF_4_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+27), 0x9E1F9B5EL);
  FF_4_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+12), 0x21C66842L);
  FF_4_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ 9), 0xF6E96C9AL);

  FF_4_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ 1), 0x670C9C61L);
  FF_4_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+29), 0xABD388F0L);
  FF_4_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ 5), 0x6A51A0D2L);
  FF_4_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+15), 0xD8542F68L);
  FF_4_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+17), 0x960FA728L);
  FF_4_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+10), 0xAB5133A3L);
  FF_4_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+16), 0x6EEF0B6CL);
  FF_4_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+13), 0x137A3BE4L);

  /* Pass 5. executed only when PASS = 5 */
  FF_5_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+27), 0xBA3BF050L);
  FF_5_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 3), 0x7EFB2A98L);
  FF_5_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+21), 0xA1F1651DL);
  FF_5_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+26), 0x39AF0176L);
  FF_5_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+17), 0x66CA593EL);
  FF_5_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+11), 0x82430E88L);
  FF_5_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+20), 0x8CEE8619L);
  FF_5_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+29), 0x456F9FB4L);

  FF_5_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+19), 0x7D84A5C3L);
  FF_5_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w   ), 0x3B8B5EBEL);
  FF_5_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+12), 0xE06F75D8L);
  FF_5_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ 7), 0x85C12073L);
  FF_5_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+13), 0x401A449FL);
  FF_5_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ 8), 0x56C16AA6L);
  FF_5_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+31), 0x4ED3AA62L);
  FF_5_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+10), 0x363F7706L);

  FF_5_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ 5), 0x1BFEDF72L);
  FF_5_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ 9), 0x429B023DL);
  FF_5_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+14), 0x37D0D724L);
  FF_5_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+30), 0xD00A1248L);
  FF_5_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+18), 0xDB0FEAD3L);
  FF_5_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ 6), 0x49F1C09BL);
  FF_5_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+28), 0x075372C9L);
  FF_5_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+24), 0x80991B7BL);

  FF_5_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ 2), 0x25D479D8L);
  FF_5_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+23), 0xF6E8DEF7L);
  FF_5_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+16), 0xE3FE501AL);
  FF_5_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+22), 0xB6794C3BL);
  FF_5_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ 4), 0x976CE0BDL);
  FF_5_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ 1), 0x04C006BAL);
  FF_5_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+25), 0xC1A94FB6L);
  FF_5_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+15), 0x409F60C4L);
}

#else /* HAVAL_LOOP */

int ii, ij;

if (state->pass == 3) {
  /* Pass 1 */
 for (ii=ij=0;ii<4;ii++) {
  FF_1_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ ij)); ij++;
  FF_1_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ ij)); ij++;
  FF_1_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ ij)); ij++;
  FF_1_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ ij)); ij++;
  FF_1_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ ij)); ij++;
  FF_1_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ ij)); ij++;
  FF_1_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ ij)); ij++;
  FF_1_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ ij)); ij++;
 }
  /* Pass 2 */
 for (ii=ij=0;ii<4;ii++) {
  FF_2_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ p2a[ij]), p2b[ij]); ij++;
 }

  /* Pass 3 */
 for (ii=ij=0;ii<4;ii++) {
  FF_3_3(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_3(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_3(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_3(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_3(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_3(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_3(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_3(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ p3a[ij]), p3b[ij]); ij++;
 }
}

if (state->pass == 4) {
  /* Pass 1 */
 for (ii=ij=0;ii<4;ii++) {
  FF_1_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ ij)); ij++;
  FF_1_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ ij)); ij++;
  FF_1_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ ij)); ij++;
  FF_1_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ ij)); ij++;
  FF_1_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ ij)); ij++;
  FF_1_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ ij)); ij++;
  FF_1_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ ij)); ij++;
  FF_1_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ ij)); ij++;
 }

  /* Pass 2 */
 for (ii=ij=0;ii<4;ii++) {
  FF_2_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ p2a[ij]), p2b[ij]); ij++;
 }

  /* Pass 3 */
 for (ii=ij=0;ii<4;ii++) {
  FF_3_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ p3a[ij]), p3b[ij]); ij++;
 }

  /* Pass 4. executed only when PASS =4 or 5 */
 for (ii=ij=0;ii<4;ii++) {
  FF_4_4(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ p4a[ij]), p4b[ij]); ij++;
  FF_4_4(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ p4a[ij]), p4b[ij]); ij++;
  FF_4_4(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ p4a[ij]), p4b[ij]); ij++;
  FF_4_4(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ p4a[ij]), p4b[ij]); ij++;
  FF_4_4(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ p4a[ij]), p4b[ij]); ij++;
  FF_4_4(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ p4a[ij]), p4b[ij]); ij++;
  FF_4_4(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ p4a[ij]), p4b[ij]); ij++;
  FF_4_4(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ p4a[ij]), p4b[ij]); ij++;
 }
}

if (state->pass == 5) {
  /* Pass 1 */
 for (ii=ij=0;ii<4;ii++) {
  FF_1_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ ij)); ij++;
  FF_1_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ ij)); ij++;
  FF_1_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ ij)); ij++;
  FF_1_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ ij)); ij++;
  FF_1_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ ij)); ij++;
  FF_1_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ ij)); ij++;
  FF_1_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ ij)); ij++;
  FF_1_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ ij)); ij++;
 }

  /* Pass 2 */
 for (ii=ij=0;ii<4;ii++) {
  FF_2_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ p2a[ij]), p2b[ij]); ij++;
  FF_2_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ p2a[ij]), p2b[ij]); ij++;
 }

  /* Pass 3 */
 for (ii=ij=0;ii<4;ii++) {
  FF_3_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ p3a[ij]), p3b[ij]); ij++;
  FF_3_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ p3a[ij]), p3b[ij]); ij++;
 }

  /* Pass 4. executed only when PASS =4 or 5 */
 for (ii=ij=0;ii<4;ii++) {
  FF_4_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ p4a[ij]), p4b[ij]); ij++;
  FF_4_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ p4a[ij]), p4b[ij]); ij++;
  FF_4_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ p4a[ij]), p4b[ij]); ij++;
  FF_4_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ p4a[ij]), p4b[ij]); ij++;
  FF_4_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ p4a[ij]), p4b[ij]); ij++;
  FF_4_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ p4a[ij]), p4b[ij]); ij++;
  FF_4_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ p4a[ij]), p4b[ij]); ij++;
  FF_4_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ p4a[ij]), p4b[ij]); ij++;
 }

  /* Pass 5. executed only when PASS = 5 */
 for (ii=ij=0;ii<4;ii++) {
  FF_5_5(t7, t6, t5, t4, t3, t2, t1, t0, *(w+ p5a[ij]), p5b[ij]); ij++;
  FF_5_5(t6, t5, t4, t3, t2, t1, t0, t7, *(w+ p5a[ij]), p5b[ij]); ij++;
  FF_5_5(t5, t4, t3, t2, t1, t0, t7, t6, *(w+ p5a[ij]), p5b[ij]); ij++;
  FF_5_5(t4, t3, t2, t1, t0, t7, t6, t5, *(w+ p5a[ij]), p5b[ij]); ij++;
  FF_5_5(t3, t2, t1, t0, t7, t6, t5, t4, *(w+ p5a[ij]), p5b[ij]); ij++;
  FF_5_5(t2, t1, t0, t7, t6, t5, t4, t3, *(w+ p5a[ij]), p5b[ij]); ij++;
  FF_5_5(t1, t0, t7, t6, t5, t4, t3, t2, *(w+ p5a[ij]), p5b[ij]); ij++;
  FF_5_5(t0, t7, t6, t5, t4, t3, t2, t1, *(w+ p5a[ij]), p5b[ij]); ij++;
 }
}

#endif /* HAVAL_LOOP */

  state->fingerprint[0] += t0;
  state->fingerprint[1] += t1;
  state->fingerprint[2] += t2;
  state->fingerprint[3] += t3;
  state->fingerprint[4] += t4;
  state->fingerprint[5] += t5;
  state->fingerprint[6] += t6;
  state->fingerprint[7] += t7;
} /* haval_hash_block */

/* tailor the last output */
static void haval_tailor (haval_state *state)
{
  haval_word temp;

if (state->fptlen == 128) {
  temp = (state->fingerprint[7] & 0x000000FFL) | 
         (state->fingerprint[6] & 0xFF000000L) | 
         (state->fingerprint[5] & 0x00FF0000L) | 
         (state->fingerprint[4] & 0x0000FF00L);
  state->fingerprint[0] += rotate_right(temp,  8);

  temp = (state->fingerprint[7] & 0x0000FF00L) | 
         (state->fingerprint[6] & 0x000000FFL) | 
         (state->fingerprint[5] & 0xFF000000L) | 
         (state->fingerprint[4] & 0x00FF0000L);
  state->fingerprint[1] += rotate_right(temp, 16);

  temp  = (state->fingerprint[7] & 0x00FF0000L) | 
          (state->fingerprint[6] & 0x0000FF00L) | 
          (state->fingerprint[5] & 0x000000FFL) | 
          (state->fingerprint[4] & 0xFF000000L);
  state->fingerprint[2] += rotate_right(temp, 24);

  temp = (state->fingerprint[7] & 0xFF000000L) | 
         (state->fingerprint[6] & 0x00FF0000L) | 
         (state->fingerprint[5] & 0x0000FF00L) | 
         (state->fingerprint[4] & 0x000000FFL);
  state->fingerprint[3] += temp;
}

else if (state->fptlen == 160) {
  temp = (state->fingerprint[7] &  (haval_word)0x3F) | 
         (state->fingerprint[6] & ((haval_word)0x7F << 25)) |  
         (state->fingerprint[5] & ((haval_word)0x3F << 19));
  state->fingerprint[0] += rotate_right(temp, 19);

  temp = (state->fingerprint[7] & ((haval_word)0x3F <<  6)) | 
         (state->fingerprint[6] &  (haval_word)0x3F) |  
         (state->fingerprint[5] & ((haval_word)0x7F << 25));
  state->fingerprint[1] += rotate_right(temp, 25);

  temp = (state->fingerprint[7] & ((haval_word)0x7F << 12)) | 
         (state->fingerprint[6] & ((haval_word)0x3F <<  6)) |  
         (state->fingerprint[5] &  (haval_word)0x3F);
  state->fingerprint[2] += temp;

  temp = (state->fingerprint[7] & ((haval_word)0x3F << 19)) | 
         (state->fingerprint[6] & ((haval_word)0x7F << 12)) |  
         (state->fingerprint[5] & ((haval_word)0x3F <<  6));
  state->fingerprint[3] += temp >> 6; 

  temp = (state->fingerprint[7] & ((haval_word)0x7F << 25)) | 
         (state->fingerprint[6] & ((haval_word)0x3F << 19)) |  
         (state->fingerprint[5] & ((haval_word)0x7F << 12));
  state->fingerprint[4] += temp >> 12;
}

else if (state->fptlen == 192) {
  temp = (state->fingerprint[7] &  (haval_word)0x1F) | 
         (state->fingerprint[6] & ((haval_word)0x3F << 26));
  state->fingerprint[0] += rotate_right(temp, 26);

  temp = (state->fingerprint[7] & ((haval_word)0x1F <<  5)) | 
         (state->fingerprint[6] &  (haval_word)0x1F);
  state->fingerprint[1] += temp;

  temp = (state->fingerprint[7] & ((haval_word)0x3F << 10)) | 
         (state->fingerprint[6] & ((haval_word)0x1F <<  5));
  state->fingerprint[2] += temp >> 5;

  temp = (state->fingerprint[7] & ((haval_word)0x1F << 16)) | 
         (state->fingerprint[6] & ((haval_word)0x3F << 10));
  state->fingerprint[3] += temp >> 10;

  temp = (state->fingerprint[7] & ((haval_word)0x1F << 21)) | 
         (state->fingerprint[6] & ((haval_word)0x1F << 16));
  state->fingerprint[4] += temp >> 16;

  temp = (state->fingerprint[7] & ((haval_word)0x3F << 26)) | 
         (state->fingerprint[6] & ((haval_word)0x1F << 21));
  state->fingerprint[5] += temp >> 21;
}

else if (state->fptlen == 224) {
  state->fingerprint[0] += (state->fingerprint[7] >> 27) & 0x1F;
  state->fingerprint[1] += (state->fingerprint[7] >> 22) & 0x1F;
  state->fingerprint[2] += (state->fingerprint[7] >> 18) & 0x0F;
  state->fingerprint[3] += (state->fingerprint[7] >> 13) & 0x1F;
  state->fingerprint[4] += (state->fingerprint[7] >>  9) & 0x0F;
  state->fingerprint[5] += (state->fingerprint[7] >>  4) & 0x1F;
  state->fingerprint[6] +=  state->fingerprint[7]        & 0x0F;
}
} /* haval_tailor */

